import {IBackendConnector} from "./IBackendConnector.ts";
import {CustomPlantPage, CustomPlantPageSchema, Plant, PlantSchema, SearchRequest} from "./schemas.ts";
import {ZodError} from "zod";
import {FetchResult} from "./FetchResult.ts";

export class BackendConnectorImpl implements IBackendConnector {
    baseUrl: string;

    constructor(baseUrl: string) {
        this.baseUrl = baseUrl;
    }

    async getAllPlants(page?: number, pageSize?: number): Promise<FetchResult<CustomPlantPage>> {
        let result: FetchResult<CustomPlantPage> = {};
        try {
            const pageParam = page ? `page=${page}` : "";
            const pageSizeParam = pageSize ? `pageSize=${pageSize}` : "";
            const response = await this.makeFetchWithAuth(this.baseUrl + "/plants?" + pageParam + pageSizeParam);
            const json = await response.json();
            result = this.parsePlantsFromPage(json);
        } catch (e) {
            console.error("An error occurred while fetching all plants")
            console.error(e);
            result = {...result, err: e};
        }
        return result;
    }

    async searchPlants(search: SearchRequest): Promise<FetchResult<Plant[]>> {
        let result: FetchResult<Plant[]> = {};
        try {
            const searchParams = new URLSearchParams(
                Object.entries(search)
                    .map(([key, value]) => [key, value as string]));
            const response = await this.makeFetchWithAuth(this.baseUrl + "/plants/search?" + searchParams);
            const json = await response.json();
            result = this.parsePlants(json);
        } catch (e) {
            console.error("An error occurred while searching through plants")
            console.error(e);
            result = {...result, err: e};
        }
        return result;
    }

    async getRandomPlants(amount: number): Promise<FetchResult<Plant[]>> {
        let result: FetchResult<Plant[]> = {};
        try {
            const response = await this.makeFetchWithAuth(this.baseUrl + "/plants/randomPlants/" + amount);
            const json = await response.json();
            result = this.parsePlants(json);
        } catch (e) {
            console.error("An error occurred while getting random plants")
            console.error(e);
            result = {...result, err: e};
        }
        return result;
    }

    async getSinglePlant(id: number): Promise<FetchResult<Plant | null>> {
        let result: FetchResult<Plant> = {};
        try {
            const response = await this.makeFetchWithAuth(this.baseUrl + "/plants/" + id);
            const json = await response.json();
            result = this.parseSinglePlant(json);
        } catch (e) {
            console.error("An error occurred while fetching plant with ID", id);
            console.error(e);
            result = {...result, err: e};
        }
        return result;
    }

    async addToGarden(plantId: number): Promise<FetchResult<void>> {
        let result: FetchResult<void> = {};
        try {
            await this.makeFetchWithAuthAndOptions(this.baseUrl + "/garden/add" + plantId, {method: "POST"});
        } catch (e) {
            console.error("An error occurred while adding plant with ID", plantId, "to garden");
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async clearGarden(): Promise<FetchResult<void>> {
        let result: FetchResult<void> = {};
        try {
            await this.makeFetchWithAuthAndOptions(this.baseUrl + "/garden/remove/all", {method: "DELETE"});
        } catch (e) {
            console.error("An error occurred while clearing garden");
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    async getGardenEntries(sort?: string): Promise<FetchResult<Plant[]>> {
        let result: FetchResult<Plant[]> = {};
        try {
            const params = new URLSearchParams();
            if (sort) params.set("sort", sort);
            const response = await this.makeFetchWithAuth(this.baseUrl + "/garden" + params);
            const json = await response.json();
            result = this.parsePlants(json);
        } catch (e) {
            console.error("An error occurred while getting garden entries")
            console.error(e);
            result = {...result, err: e};
        }
        return result;
    }

    async getGardenSize(): Promise<FetchResult<number>> {
        let result: FetchResult<number> = {};
        try {
            const response = await this.makeFetchWithAuth(this.baseUrl + "/garden/count");
            const json = await response.json();
            const size = Number.parseInt(json);
            result = {value: size};
        } catch (e) {
            console.error("An error occurred while fetching garden size")
            console.error(e);
            result = {...result, err: e};
        }
        return result;
    }

    async removeEntryFromGarden(entryId: number): Promise<FetchResult<void>> {
        let result: FetchResult<void> = {};
        try {
            await this.makeFetchWithAuthAndOptions(this.baseUrl + "/garden/remove/" + entryId, {method: "DELETE"});
        } catch (e) {
            console.error("An error occurred while removing garden entry with ID", entryId);
            console.error(e);
            result = {err: e};
        }
        return result;
    }

    makeFetchWithAuth(fullUrl: string): Promise<Response> {
        return this.makeFetchWithAuthAndOptions(fullUrl, {});
    }

    makeFetchWithAuthAndOptions(fullUrl: string, options: {
        method?: "GET" | "POST" | "DELETE" | "PUT" | "PATCH",
        body?: object
    }): Promise<Response> {
        //TODO: get auth
        console.log("Making request to", fullUrl)
        const {method, body} = options;
        const jwt: string = "eyJhbGciOiJIUzI1NiJ9.eyJzdWIiOiJsdWthcy5rYXJzY2hAZ214LmRlIiwiaWF0IjoxNzAzMjYwNDU5LCJleHAiOjE3MDMzNDY4NTl9.Crt0lc8CgpQAQdi_--9CLZxpaLcb_1IM22-aRAbbMuU";
        let headers: HeadersInit = {
            Authorization: `Bearer ${jwt}`
        };
        if (body) {
            headers = {...headers, contentType: "application/json"}
        }
        return fetch(fullUrl, {
            method: method,
            headers: headers,
            body: JSON.stringify(body),
        });
    }

    parseSinglePlant(json: any): FetchResult<Plant> {
        let result: FetchResult<Plant> = {};
        try {
            result = {value: PlantSchema.parse(json)};
        } catch (e) {
            if (e instanceof ZodError) {
                result = {...result, err: "One or more plants could not be parsed correctly."}
            }
        }
        return result;
    }

    parsePlants(json: any): FetchResult<Plant[]> {
        let result: FetchResult<Plant[]> = {};
        const plants: Plant[] = [];
        for (const p of json) {
            try {
                plants.push(PlantSchema.parse(p))
            } catch (e) {
                if (e instanceof ZodError) {
                    result = {...result, err: "One or more plants could not be parsed correctly."}
                }
            }
        }
        return {...result, value: plants};
    }

    parsePlantsFromPage(json: any): FetchResult<CustomPlantPage> {
        let result: FetchResult<CustomPlantPage> = {};
        try {
            result = {value: CustomPlantPageSchema.parse(json)}
        } catch (e) {
            if (e instanceof ZodError) {
                result = {...result, err: "One or more plants could not be parsed correctly."}
            }
        }
        return result;
    }
}
